
#include <stdio.h>
#include <stdlib.h>
//#include <unistd.h>
#include <string.h>
#include "cell.h"
#include "sif.h"

typedef unsigned long	uLong;
typedef unsigned short	uShort;

static FILE	*tiff;
static long	putBytes;

#ifdef	LZW
static int	numBits,curBits;
static long	curData;

static void	Put(code)
int		code;
{
	curData=(curData<<numBits)|(long)code; curBits+=numBits;
	while (curBits>=8) {
	    curBits-=8; fputc((int)((curData>>curBits)&0xffL),tiff); ++putBytes;
	}
}

#define MINBITS	9
#define MAXBITS	12
#define CLRCODE	(1<<(MINBITS-1))
#define EODCODE	(CLRCODE+1)
#define MINCODE	(CLRCODE+2)
#define MAXCODE	((1<<MAXBITS)-2)

#define TBLSIZE	9001
#define SHIFT	5
#define EMPTY	0xffffffffL

static int	oldCode,limCode,entCode,codeTbl[TBLSIZE];
static long	keyTbl[TBLSIZE];

static void	InitEncode()
{
	int	idx;

	oldCode=(-1); curBits=0; putBytes=0L;

	for (idx=0; idx<TBLSIZE; idx++) keyTbl[idx]=EMPTY;
	numBits=MINBITS; limCode=(1<<MINBITS)-1; entCode=MINCODE;
	Put(CLRCODE);
}

static long	TermEncode()
{
	if (oldCode>=0) Put(oldCode);

	Put(EODCODE);

	if (curBits>0) {
	    fputc((int)((curData<<(8-curBits))&0xffL),tiff); ++putBytes;
	}
	return putBytes;
}

static void	Encode(newCode)
int		newCode;
{
	int	idx,step;
	long	key;

	if (oldCode<0) {
	    oldCode=newCode; return;
	}

	key=(long)oldCode|((long)newCode<<MAXBITS);
	idx=(newCode<<SHIFT)^oldCode;
	if (keyTbl[idx]==key) {
	    oldCode=codeTbl[idx]; return;
	}
	if (keyTbl[idx]!=EMPTY) {
	    step=(idx==0)?1:TBLSIZE-idx;
	    do {
		if ((idx-=step)<0) idx+=TBLSIZE;

		if (keyTbl[idx]==key) {
		    oldCode=codeTbl[idx]; return;
		}
	    } while (keyTbl[idx]!=EMPTY);
	}
	Put(oldCode);
	keyTbl[idx]=key; codeTbl[idx]=entCode;
	oldCode=newCode;

	if (++entCode==MAXCODE) {
	    Put(CLRCODE);
	    numBits=MINBITS; limCode=(1<<MINBITS)-1; entCode=MINCODE;
	    for (idx=0; idx<TBLSIZE; idx++) keyTbl[idx]=EMPTY;
	}
	else
	    if (entCode>limCode) {
		++numBits; limCode=(1<<numBits)-1;
	    }
}
#else
#define InitEncode()	putBytes=0L
#define TermEncode()	putBytes
#define Encode(code)	{ fputc(code,tiff); ++putBytes; }
#endif

static void	Put2(data)
uShort		data;
{
	fputc((int)((data>>8)&0xff),tiff); fputc((int)(data&0xff),tiff);
}

static void	Put4(data)
uLong		data;
{
	Put2((uShort)((data>>16)&0xffffL)); Put2((uShort)(data&0xffffL));
}

#define PutIFD(tag,type,count)	Put2(tag); Put2(type); Put4((uLong)(count))
#define PutIFD3(tag,data)	PutIFD(tag,3,1); Put2((uShort)(data)); Put2(0)
#define PutIFD4(tag,count,data)	PutIFD(tag,4,count); Put4((uLong)(data))

#ifdef	FUNCTION_CELL
#define FETCH(x,y)	(*cell)(x,y)
#else
#define FETCH(x,y)	cell[y][x]
#endif

#define ERROR(msg)	{ fprintf(stderr,"%s : %s\n",path,msg); exit(1); }

void	StoreTiffFile(path,Nx,Ny,BPS,cell,desc)
char	*path,*desc;
int	Nx,Ny,BPS;
#ifdef	FUNCTION_CELL
Cell	(*cell)();
#else
Cell	**cell;
#endif
{
	long	bpr,*cnt,ofs,ifd;
	int	rps,spi,y,x,dat1,bit1,idx;
	size_t	len;
	Cell	msb,dat2,bit2;

	if (Nx<=0 || Ny<=0 || BPS<=0 || (tiff=fopen(path,"wb"))==NULL)
	    ERROR("file not stored.");

	bpr=((long)Nx*(long)BPS+7L)/8L;
	rps=(bpr>8192L)?1:(int)(8192L/bpr);
	spi=(Ny+rps-1)/rps;
	if ((cnt=(long *)malloc((size_t)spi*sizeof(*cnt)))==NULL) {
	    fclose(tiff); unlink(path);
	    ERROR("TIFF work memory allocation error.");
	}

	fseek(tiff,8L,0);

	len=(desc!=NULL)?strlen(desc)+1:0;
	if (len>4) fwrite(desc,len+(len&1),1,tiff);

	ofs=ftell(tiff);

	if (BPS!=8 && BPS!=16) msb=(Cell)(1<<(BPS-1));

	for (y=0; y<Ny; y++) {
	    if (y%rps==0) InitEncode();

	    if (BPS==8)
		for (x=0; x<Nx; x++) {
		    Encode((int)FETCH(x,y));
		}
	    else if (BPS==16)
		for (x=0; x<Nx; x++) {
		    dat2=FETCH(x,y);
		    Encode((int)((dat2>>8)&0xff));
		    Encode((int)( dat2    &0xff));
		}
	    else {
		dat1=0; bit1=0x80;
		for (x=0; x<Nx; x++) {
		    dat2=FETCH(x,y);
		    for (bit2=msb; bit2!=0; bit2>>=1) {
			if (dat2&bit2) dat1|=bit1;
			if (bit1==0x01) {
			    Encode(dat1); dat1=0; bit1=0x80;
			}
			else
			    bit1>>=1;
		    }
		}
		if (bit1!=0x80) Encode(dat1);
	    }
	    if (y+1==Ny || (y+1)%rps==0) cnt[y/rps]=TermEncode();
	}
	if (ftell(tiff)&1L) fputc(0,tiff);

	if (spi>1) {
	    for (idx=0; idx<spi; idx++) {
		Put4((uLong)ofs); ofs+=cnt[idx];
	    }
	    if (ofs&1L) ++ofs;

	    for (idx=0; idx<spi; idx++) Put4((uLong)cnt[idx]);
	}
	ifd=ftell(tiff);

	Put2((len)?12:11);		/* number of tags */

	PutIFD3(0x100,Nx);		/* 0: image width */
	PutIFD3(0x101,Ny);		/* 1: image length */
	PutIFD3(0x102,BPS);		/* 2: bits per sample */
#ifdef	LZW
	PutIFD3(0x103,5);		/* 3: image compression */
#else
	PutIFD3(0x103,1);		/* 3: image compression */
#endif
	PutIFD3(0x106,(BPS==1)?0:1);	/* 4: photometric interpretation */
	PutIFD3(0x10a,1);		/* 5: fill order */
	if (len) {
	    PutIFD(0x10e,2,len);
	    if (len>4) Put4(8L); else fwrite(desc,4,1,tiff);
					/* 6: image description */
	}
	PutIFD4(0x111,spi,ofs); 	/* 6/7: strip offset */
	PutIFD3(0x115,1);		/* 7/8: samples per pixel */
	PutIFD3(0x116,rps);		/* 8/9: rows per strip */
	PutIFD4(0x117,spi,(spi>1)?ofs+(long)spi*4L:*cnt);
					/* 9/10: strip byte counts */
	PutIFD3(0x11c,1);		/* 10/11: planar configuration */
	Put4(0L);

	fseek(tiff,0L,0);
	Put4(0x4d4d002aL);		/* magic number of TIFF */
	Put4((uLong)ifd);

	free(cnt); fclose(tiff);
}
